/*
 * Implementation of the update function
 *
 * Copyright (C) 2003  Enrico Zini <enrico@debian.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Update.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#else
#warning No config.h found: using fallback values
#define WGET "/usr/bin/wget"
#endif

#include <stringf.h>

#include "Exec.h"
#include "ChildProcess.h"
#include "Environment.h"

#include <ctype.h>

#include <unistd.h>		/* fork */
#include <sys/types.h>	/* wait, fork */
#include <sys/wait.h>	/* wait */
#include <stdlib.h>		/* getenv */
#include <errno.h>		/* errno */
 
using namespace std;
using namespace stringf;

extern const char* fn_tagdb;
extern const char* fn_impls;
extern const char* fn_deriv;
extern const char* fn_vocab;
extern const char* fn_dist_vocab;
extern const char* fn_sources;
extern const char* path_cache;
extern const char* path_tagpatch_d;
extern const char* path_tagvoc_d;
extern const char* remote_tags_fname;

static bool acquireWget(const string& url, const string& target)
	throw (SystemException)
{
	string tmptarget = target + ".partial";
	pid_t pid;
	if ((pid = fork()) == 0)
	{
		try {
			Exec prog(WGET);
			prog.addArg(WGET);
			if (!isatty(1))
				prog.addArg("-q");
			prog.addArg("-O");
			prog.addArg(tmptarget);
			prog.addArg(url);
			if (char* hp = getenv("http_proxy"))
				prog.addEnv(string("http_proxy=") + hp);
			prog.exec();
			_exit(1);
		} catch (Exception& e) {
			fatal_error("%s: %.*s\n", e.type(), PFSTR(e.desc()));
		}
	}
	if (pid == -1)
		throw SystemException(errno, "creating a child process to run wget");
	int status;
	if (wait(&status) == -1)
		throw SystemException(errno, "waiting for child completion");
	if (WIFEXITED(status) && WEXITSTATUS(status) == 0)
	{
		if (rename(tmptarget.c_str(), target.c_str()) == -1)
			throw FileException(errno, "renaming " + tmptarget + " to " + target);
		return true;
	}
	else
	{
		string fmtstatus = Process::formatStatus(status);
		warning("wget \"%.*s\" %.*s\n", PFSTR(url), PFSTR(fmtstatus));
		return false;
	}
}

static bool acquireLocal(const string& file, const string& target)
	throw (SystemException)
{
	// FIXME: this is an inefficient method of copying files, but a quick one
	// to code.  If the files to be copied get big, this function can be
	// rewritten to use open and read/write on a memory buffer

	FILE* in = fopen(file.c_str(), "rt");
	if (!in) throw FileException(errno, "opening " + file);
	string tmptarget = target + ".partial";

	FILE* out = fopen(tmptarget.c_str(), "wt");
	if (!out) throw FileException(errno, "opening " + tmptarget);
	
	int c;
	while ((c = fgetc(in)) != EOF)
		fputc(c, out);

	fclose(in);
	fclose(out);
	
	if (rename(tmptarget.c_str(), target.c_str()) == -1)
		throw FileException(errno, "renaming " + tmptarget + " to " + target);
	return true;
}

string acquireSource(const string& source) throw (Exception)
{
	string localName = string(path_cache) + '/';

	// Generate the local name by normalization of the source uri
	for (string::const_iterator s = source.begin();
			s != source.end(); s++)
		if (isalnum(*s) || *s == '-')
			localName += *s;
		else
			localName += '_';
	
	if (source.substr(0, 7) == "http://" ||
		source.substr(0, 6) == "ftp://")
	{
		acquireWget(source + "/" + remote_tags_fname, localName);
	}
	else if (source.substr(0, 5) == "file:")
	{
		acquireLocal(source.substr(5), localName);
	}
	else
	{
		warning("Source %.*s as an unsupported scheme\n", PFSTR(source));
	}

	return localName;
}


// vim:set ts=4 sw=4:
