/*
 * Tag vocabulary access
 *
 * Copyright (C) 2003  Enrico Zini <enrico@debian.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Vocabulary.h"

#include "DebDBParser.h"
#include "StdioParserInput.h"
#include "StringParserInput.h"
#include "Environment.h"

#include <stringf.h>
#include <Tagexpr.h>
#include <TagexprParser.h>
#include <TagcollParser.h>

#include <sys/types.h>  // opendir, readdir, closedir
#include <dirent.h>     // opendir, readdir, closedir
#include <unistd.h>     // access
#include <errno.h>

using namespace std;

extern const char* fn_tagdb;
extern const char* fn_impls;
extern const char* fn_deriv;
extern const char* fn_vocab;
extern const char* fn_dist_vocab;
extern const char* fn_sources;
extern const char* path_cache;
extern const char* path_tagpatch_d;
extern const char* path_tagvoc_d;

static void writeDebStyleField(FILE* out, const string& name, const string& val) throw ()
{
	fprintf(out, "%.*s: ", PFSTR(name));

	// Properly escape newlines
	bool was_nl = false;
	for (string::const_iterator s = val.begin(); s != val.end(); s++)
		if (was_nl)
			// \n\n -> \n .\n
			if (*s == '\n')
			{
				fputc(' ', out);
				fputc('.', out);
				fputc(*s, out);
			}
			// \n([^ \t]) -> \n \1
			else if (*s != ' ' && *s != '\t')
			{
				fputc(' ', out);
				fputc(*s, out);
				was_nl = false;
			}
			// \n[ \t] goes unchanged
			else
			{
				fputc(*s, out);
				was_nl = false;
			}
		else
			if (*s == '\n')
			{
				fputc(*s, out);
				was_nl = true;
			}
			else
				fputc(*s, out);

	fputc('\n', out);
}

void Vocabulary::Record::parseImplies(const string& impls) throw (ParserException)
{
	StringParserInput in(impls);
	string item;
	int sep;
	do
	{
		sep = TagcollParser::parseElement(in, item);
		if (item.size())
			implies.push_back(item);
	}
	while (sep != ParserInput::Eof);
}

string Vocabulary::Record::formatImplies() const throw ()
{
	string res;

	for (list<string>::const_iterator i = implies.begin();
			i != implies.end(); i++)
		if (i == implies.begin())
			res += *i;
		else
			res += ", " + *i;

	return res;
}


void Vocabulary::read(ParserInput& input) throw (ParserException)
{
	DebDBParser parser(input);
	DebDBParser::Record rec;
	TagexprParser teparser;

	while (parser.nextRecord(rec))
	{
		string name = rec["Tag"];
		if (name.size())
		{
			Record r;
			for (DebDBParser::Record::const_iterator i = rec.begin();
					i != rec.end(); i++)
			{
				if (i->first == "Tag")
					;
				else if (i->first == "Implies")
					r.parseImplies(i->second);
				else if (i->first == "Derived")
					r.derived = teparser.parse(i->second);
				else if (i->first == "Description")
					r.description = i->second;
				else
				{
					verbose("%.*s:%d: unrecognized vocabulary mapping: \"%.*s: %.*s\"\n",
							PFSTR(input.fileName()), input.lineNumber(),
							PFSTR(i->first), PFSTR(i->second));
					r.others.push_back(make_pair(i->first, i->second));
				}
			}
			contents.insert(make_pair(name, r));
		} else {
			warning("%.*s:%d: Skipping record without Name: field\n",
					PFSTR(input.fileName()), input.lineNumber());
		}
		
		rec.clear();
	}
}

void Vocabulary::read() throw (FileException, ParserException)
{
	StdioParserInput mainInput(fn_vocab);
	read(mainInput);
}

void Vocabulary::readVocabularies() throw (SystemException, FileException, ParserException)
{
	StdioParserInput mainInput(fn_dist_vocab);
	read(mainInput);
	
	DIR* dir = opendir(path_tagvoc_d);
	if (!dir)
		throw SystemException(errno, string("reading directory ") + path_tagvoc_d);

	while (struct dirent* d = readdir(dir))
	{
		if (d->d_name[0] != '.' &&
				d->d_name[strlen(d->d_name) - 1] != '~')
		{
			string fn = string(path_tagvoc_d) + '/' + d->d_name;
			if (access(fn.c_str(), R_OK) == 0)
			{
				StdioParserInput patchInput(fn);
				read(patchInput);
			}
		}
	}
	closedir(dir);
}

void Vocabulary::write() throw (FileException)
{
	string fn = string(fn_vocab) + ".tmp";
	FILE* out = fopen(fn.c_str(), "wt");
	if (!out) throw FileException(errno, "creating file " + fn);

	for (const_iterator i = begin(); i != end(); i++)
	{
		writeDebStyleField(out, "Tag", i->first);
		if (i->second.implies.size())
			writeDebStyleField(out, "Implies", i->second.formatImplies());
		if (i->second.derived)
		{
			Tagexpr* e = i->second.derived;
			writeDebStyleField(out, "Derived", e->format());
			//writeDebStyleField(out, "Derived", i->second.derived->format());
		}
		if (i->second.description.size())
			writeDebStyleField(out, "Description", i->second.description);

		for (list< pair<string, string> >::const_iterator j = i->second.others.begin();
				j != i->second.others.end(); j++)
			if (j->second.size())
				writeDebStyleField(out, j->first, j->second);

		fputc('\n', out);
	}

	fclose(out);

	if (rename(fn.c_str(), fn_vocab) == -1)
		throw FileException(errno, "renaming " + fn + " to " + fn_vocab);
}

void Vocabulary::writeImplications() throw (FileException)
{
	string fn = string(fn_impls) + ".tmp";
	FILE* out = fopen(fn.c_str(), "wt");
	if (!out) throw FileException(errno, "creating file " + fn);

	for (const_iterator i = begin(); i != end(); i++)
		if (i->second.implies.size())
		{
			string formatted = i->second.formatImplies();
			fprintf(out, "%.*s: %.*s\n", PFSTR(i->first), PFSTR(formatted));
		}

	fclose(out);

	if (rename(fn.c_str(), fn_impls) == -1)
		throw FileException(errno, "renaming " + fn + " to " + fn_impls);
}

void Vocabulary::writeDerivedTags() throw (FileException)
{
	string fn = string(fn_deriv) + ".tmp";
	FILE* out = fopen(fn.c_str(), "wt");
	if (!out) throw FileException(errno, "creating file " + fn);

	for (const_iterator i = begin(); i != end(); i++)
		if (i->second.derived)
		{
			string formatted = i->second.derived->format();
			fprintf(out, "%.*s: %.*s\n", PFSTR(i->first), PFSTR(formatted));
		}

	fclose(out);

	if (rename(fn.c_str(), fn_deriv) == -1)
		throw FileException(errno, "renaming " + fn + " to " + fn_deriv);
}

void Vocabulary::outputImplications(TagcollConsumer<string>& consumer) const throw ()
{
	for (const_iterator i = begin(); i != end(); i++)
		if (!i->second.implies.empty())
		{
			OpSet<string> ts;
			copy(i->second.implies.begin(), i->second.implies.end(), inserter(ts, ts.begin()));
			consumer.consume(i->first, ts);
		}
}

void Vocabulary::outputDerivedTags(DerivedTagList& dtlist) const throw ()
{
	for (const_iterator i = begin(); i != end(); i++)
		if (i->second.derived)
			dtlist.add(i->first, i->second.derived);
}

// vim:set ts=4 sw=4:
