
/*
 * debtags - Implement package tags support for Debian
 *
 * Copyright (C) 2003  Enrico Zini <enrico@debian.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#define APPNAME PACKAGE_NAME
#else
#warning No config.h found: using fallback values
#define APPNAME __FILE__
#define PACKAGE_VERSION "unknown"
#endif

#include <FilterChain.h>
#include <Implications.h>
#include <TagcollFilter.h>
#include <DerivedTags.h>
#include <InputMerger.h>
#include <TagcollSerializer.h>
#include <TagcollParser.h>
#include <ItemGrouper.h>
#include <ExpressionFilter.h>
#include <HandleMaker.h>
#include <TagCollection.h>
#include <TagcollBuilder.h>
#include <Patches.h>

#include <Exception.h>
#include <stringf.h>

#include <list>

#include <cctype>

#include <errno.h>

#include "Environment.h"
#include "CommandlineParser.h"
#include "TagcollSink.h"
#include "DebDBParser.h"
#include "StdioParserInput.h"
#include "StringParserInput.h"
#include "ZlibParserInput.h"
#include "PackageDB.h"
#include "Vocabulary.h"
#include "Update.h"

#include <sys/types.h>  // opendir, readdir, closedir
#include <dirent.h>     // opendir, readdir, closedir
#include <unistd.h>     // access

using namespace std;


const char* fn_tagdb = "/var/lib/debtags/package-tags";
const char* fn_impls = "/var/lib/debtags/implications";
const char* fn_deriv = "/var/lib/debtags/derived-tags";
const char* fn_vocab = "/var/lib/debtags/vocabulary";
const char* fn_dist_vocab = "/usr/share/debtags/vocabulary";
const char* fn_sources = "/etc/debtags/sources.list";
const char* path_cache = "/var/cache/debtags";
const char* path_tagpatch_d = "/etc/debtags/tagpatch.d";	
const char* path_tagvoc_d = "/etc/debtags/tagvoc.d";	
const char* remote_tags_fname = "tags-current.gz";

void wantTagDatabase() throw ()
{
	if (   access(fn_tagdb, R_OK) == -1
		|| access(fn_impls, R_OK) == -1
		|| access(fn_deriv, R_OK) == -1
		|| access(fn_vocab, R_OK) == -1)
		fatal_error(
				"The tag database has not yet been generated: you need to run \"debtags update\" (as root) before using the package tags\n");
}

void readCollection(const string& file, TagcollConsumer<string>& builder)
	throw (FileException, ParserException)
{
	if (file == "-")
	{
		StdioParserInput input(stdin, "<stdin>");
		TagcollParser::parseTagcoll(input, builder);
	}
	else
	{
		StdioParserInput input(file);
		TagcollParser::parseTagcoll(input, builder);
	}
}

void zreadCollection(const string& file, TagcollConsumer<string>& builder)
	throw (FileException, ParserException)
{
	ZlibParserInput in(file);

	// Read the collection
	TagcollParser::parseTagcoll(in, builder);
}

void readDerivedTags(const string& file, DerivedTagList& derivedTags)
	throw (FileException, ParserException)
{
	if (file == "-")
	{
		StdioParserInput input(stdin, "<stdin>");
		derivedTags.parse(input);
	}
	else
	{
		StdioParserInput input(file);
		derivedTags.parse(input);
	}
}

void readFullCollection(const std::string& file, TagcollConsumer<string>& consumer)
{
	// Prepare the input filter chain
	FilterChain<string> filters;

	// Expand implications
	ImplicationList implications;
	readCollection(fn_impls, implications);
	// Pack the structure for faster expansions
	implications.pack();
	TagcollFilter<string>* ximpl = new ApplyImplications(implications);
	filters.appendFilter(ximpl);

	// Add derived tags
	DerivedTagList derivedTags;
	readDerivedTags(fn_deriv, derivedTags);
	TagcollFilter<string>* xderv = new ApplyDerivedTags(derivedTags);
	filters.appendFilter(xderv);

	// Add further tags implicated by the derived tags
	filters.appendFilter(ximpl);

	// Read and expand the database
	InputMerger<string> merger;
	filters.setConsumer(&consumer);
	readCollection(file, merger);

	// Output the merged collection to consumer
	merger.output(filters);

	delete ximpl;
	delete xderv;
}

void readFullCollection(TagcollConsumer<string>& consumer)
{
	readFullCollection(fn_tagdb, consumer);
}
 
void readUnmergedFullCollection(TagcollConsumer<string>& consumer)
{
	// Prepare the input filter chain
	FilterChain<string> filters;

	// Expand implications
	ImplicationList implications;
	readCollection(fn_impls, implications);
	// Pack the structure for faster expansions
	implications.pack();
	TagcollFilter<string>* ximpl = new ApplyImplications(implications);
	filters.appendFilter(ximpl);

	// Add derived tags
	DerivedTagList derivedTags;
	readDerivedTags(fn_deriv, derivedTags);
	TagcollFilter<string>* xderv = new ApplyDerivedTags(derivedTags);
	filters.appendFilter(xderv);

	// Add further tags implicated by the derived tags
	filters.appendFilter(ximpl);

	// Read and expand the database
	filters.setConsumer(&consumer);
	readCollection(fn_tagdb, filters);

	delete ximpl;
	delete xderv;
}

static void readUnmergedCollection(TagcollConsumer<string>& consumer)
{
	// Read the database
	readCollection(fn_tagdb, consumer);
}

// Catch the tags for a given item
class TagcollCatcher : public TagcollConsumer<string>
{
protected:
	string _item;
	OpSet<string> _tags;
	
public:
	TagcollCatcher(const std::string& item) throw ()
		: _item(item) {}
	virtual ~TagcollCatcher() throw () {}

	const OpSet<string>& tags() const throw () { return _tags; }

	// If the package is really untagged, there's nothing we can do here
	virtual void consume(const std::string& item) throw () {}
	virtual void consume(const OpSet<string>& items) throw () {}
	
	// Catch the tags that belong to _item
	virtual void consume(const std::string& item, const OpSet<string>& tags) throw ()
	{
		if (item == _item)
			_tags += tags;
	}

	// Catch the tags that belong to _item
	virtual void consume(const OpSet<string>& items, const OpSet<string>& tags) throw ()
	{
		if (items.contains(_item))
			_tags += tags;
	}
};

// Check a collection agaisnt a tag vocabulary
class TagcollChecker : public TagcollConsumer<string>
{
protected:
	const Vocabulary& voc;
	int missing_count;
	map<string, int> missing;
	unsigned int mtag_width;
	int max_missing;

	void checkTag(const std::string& tag) throw ()
	{
		if (!voc.has(tag))
		{
			//warning("Tag %.*s not found in vocabulary\n", PFSTR(tag));
			missing_count++;
			missing[tag]++;
			if (tag.size() > mtag_width)
				mtag_width = tag.size();
			if (missing[tag] > max_missing)
				max_missing = missing[tag];
		}
	}
	
public:
	TagcollChecker(const Vocabulary& voc) throw ()
		: voc(voc), missing_count(0), mtag_width(0), max_missing(0) {}
	virtual ~TagcollChecker() throw () {}

	// Get the number of times a bad tag has been found in input
	int missingCount() const throw () { return missing_count; }

	// If the package is really untagged, there's nothing we can do here
	virtual void consume(const std::string& item) throw () {}
	virtual void consume(const OpSet<string>& items) throw () {}
	
	// Catch the tags that belong to _item
	virtual void consume(const std::string& item, const OpSet<string>& tags) throw ()
	{
		for (OpSet<string>::const_iterator i = tags.begin(); i != tags.end(); i++)
			checkTag(*i);
	}

	// Catch the tags that belong to _item
	virtual void consume(const OpSet<string>& items, const OpSet<string>& tags) throw ()
	{
		for (OpSet<string>::const_iterator i = tags.begin(); i != tags.end(); i++)
			checkTag(*i);
	}

	void report() throw ()
	{
		if (missing_count > 0)
		{
			warning("%d missing tags found in %d occasions\n", missing.size(), missing_count);
			warning("Missing tags were:\n");

			int mtag_maxdigits = 0;
			for (int i = max_missing; i > 0; i = i / 10)
				mtag_maxdigits++;

			for (map<string, int>::const_iterator i = missing.begin();
					i != missing.end(); i++)
			{
				warning("\t%.*s %*s (%*d time%s)\n",
					PFSTR(i->first), 
					mtag_width - i->first.size(),
					"", 
					mtag_maxdigits,
					i->second,
					i->second > 1 ? "s" : "");
			}
		}
	}
};

class ItemCollector : public OpSet<string>, public TagcollConsumer<string, string>
{
public:
	// Process an untagged item
	virtual void consume(const string& item) throw () { (*this) += item; }

	// Process a tagged item, with its tags
	virtual void consume(const string& item, const OpSet<string>& tags) throw ()  { (*this) += item; }

	// Process a set of items, all with no tags
	virtual void consume(const OpSet<string>& items) throw () { (*this) += items; }

	// Process a set of items identically tagged, with their tags
	virtual void consume(const OpSet<string>& items, const OpSet<string>& tags) throw ()
		{ (*this) += items; }
};


static HandleMaker<string> handleMaker;

void printItems(PackageDB db, const set<int>& items) throw (PackageDBException, ParserException)
{
	for (set<int>::const_iterator i = items.begin();
			i != items.end(); i++)
	{
		string pkg = handleMaker.getItem(*i);
		string desc;

		try
		{
			string rec = db.getPackageRecord(pkg);
			StringParserInput input(rec);

			//debug("REC: %.*s\n", PFSTR(rec));

			DebDBParser parser(input);
			DebDBParser::Record r;
			if (!parser.nextRecord(r))
				desc = "Description for package " + pkg + " not found";
			else
			{
				desc = r["Description"];
				desc = desc.substr(0, desc.find("\n"));
			}

			printf("%.*s: %.*s\n", PFSTR(pkg), PFSTR(desc));
		}
		catch (PackageDBException&)
		{
			/* Ignore packages not available */
		}
	}
}


static void printVocabularyItem(const string& tag, const Vocabulary::Record& rec) throw ()
{
	printf("Tag: %.*s\n", PFSTR(tag));
	if (!rec.implies.empty())
	{
		string formatted = rec.formatImplies();
		printf("Implies: %.*s\n", PFSTR(formatted));
	}
	if (rec.derived)
	{
		string formatted = rec.derived->format();
		printf("Derived: %.*s\n", PFSTR(formatted));
	}
	if (!rec.description.empty())
		printf("Description: %.*s\n", PFSTR(rec.description));
	for (list< pair<string, string> >::const_iterator j = rec.others.begin();
			j != rec.others.end(); j++)
		printf("%.*s: %.*s\n", PFSTR(j->first), PFSTR(j->second));
}

static bool matchVocabularyItem(const string& tag, const Vocabulary::Record& rec, const string& pat) throw ()
{
	if (tag.find(pat) != string::npos
		|| rec.description.find(pat) != string::npos)
		return true;
	else
		for (list< pair<string, string> >::const_iterator j = rec.others.begin();
				j != rec.others.end(); j++)
			if (j->second.find(pat) != string::npos)
				return true;
	return false;
}

static vector<string> readSources() throw (FileException)
{
	FILE* in = fopen(fn_sources, "rt");
	if (!in)
		throw FileException(errno, string("reading ") + fn_sources);

	vector<string> res;

	string line;
	int c;
	while ((c = fgetc(in)) != EOF)
	{
		if (c != '\n')
			line += c;
		else
		{
			unsigned int i = 0;
	
			// Skip leading spaces
			while (i < line.size() && isspace(line[i]))
				i++;

			// If it's a tag source
			if (line.substr(i, 4) == "tags")
			{
				i += 4;
				// Skip further spaces
				while (i < line.size() && isspace(line[i]))
					i++;
				res.push_back(line.substr(i));
			}
			
			line = string();
		}
	}

	fclose(in);

	return res;
}

class CommandlineParserWithCommand : public CommandlineParser
{
protected:
	map<string, int> command_map;

public:
	CommandlineParserWithCommand(const std::string& argv0,
									const std::string& cmdline_summary,
									const std::string& description) throw ()
		: CommandlineParser(argv0, cmdline_summary, description)
	{
		add("version", 'V', "version", "print the program version, then exit");
	}

	void addCommand(const std::string name, int id) throw ()
	{
		command_map[name] = id;
	}

	int parse(int& argc, const char**& argv) throw ()
	{
		if (!CommandlineParser::parse(argc, argv))
		{
			printHelp();
			return false;
		}
		if (get("help").defined())
		{
			printHelp();
			exit(0);
		}
		if (get("version").defined())
		{
			printf("%s ver." PACKAGE_VERSION "\n", APPNAME);
			exit(0);
		}
		if (argc == 1)
		{
			printHelp();
			exit(1);
		}

		string command_string = argv[1];
		map<string, int>::const_iterator cmap_i = command_map.find(command_string);
		if (cmap_i == command_map.end())
		{
			fprintf(stderr, "Invalid command: \"%.*s\"\n", PFSTR(command_string));
			printHelp();
			exit(1);
		}

		for (int i = 1; i < argc; i++)
			argv[i] = argv[i + 1];
		--argc;
		
		return cmap_i->second;
	}
};

class CommandlineArgs
{
protected:
	int argc;
	const char** argv;
	int _next;

public:
	CommandlineArgs(int argc, const char* argv[]) throw () : argc(argc), argv(argv), _next(1) {}

	// Return true if there is another argument left in the list
	bool hasNext() const throw () { return argc >= _next + 1; }

	// Return the next argument in the list
	string next() throw ()
	{
		if (hasNext())
		{
			return argv[_next++];
		} else {
			return "-";
		}
	}
};

enum valid_command { UPDATE, CHECK, TAGSHOW, TAGSEARCH, SHOW, RELATED, CAT, GREP, INSTALL, MKPATCH };

int main(int argc, const char* argv[])
{
	try {
		CommandlineParserWithCommand opts(APPNAME, "[options] [command] [file1|-] [file2|-]",
				"Perform various operations on a tagged collection\n\n"
				"Commands are:\n"
				"  update        Updates the package tag database (requires root)\n"
				"  check <file>	 Check that all the tags in the given tagged collection are present\n"
				"                in the tag vocabulary.  Checks the main database if no file is\n"
				"                specified\n"
				"  tagshow <tag> Show the vocabulary informations about a tag\n"
				"  tagsearch <pattern [pattern [pattern [...]]]>\n"
				"                Show a summary of all tags matching the given patterns\n"
				"  show <pkg>    Call apt-cache show <pkg>, but add tag informations to the output.\n"
				"  related <pkg1[,pkg2[,pkg2,[...]]]>\n"
				"                Show packages related to the specified ones.\n"
				"  cat           Output the full package tag database\n"
				"  grep [-v] [-q] <tag expression>\n"
				"                Output the lines of the full package tag database that match\n"
				"                the given tag expression\n"
				"  install [-v] [-q] <tag expression>\n"
				"                apt-get install the packages that match the given tag expression\n"
				"  mkpatch [filename]\n"
				"                Create a tag patch between the current tag database and the tag\n"
				"                collection [filename]\n");


		/*
		opts.add("hierarchy", 's', "smart-hierarchy", "build a smart hierarchy");
		opts.add("implications", 'm', "show-implications", "output a list of tag implications");
		opts.add("copy", 'c', "copy", "output the collection");
		opts.add("diff", 'd', "diff", "output a tag patch file with the differences between two files (requires two file arguments)");
		*/

		//opts.add("expanded", 'x', "expanded-output", "produce full (and redundant) output data instead of compact");
		opts.add("groupitems", 'g', "group-items", "group items with the same tagset in the output collection");
		opts.add("distance", 'd', "distance", "set the maximum distance to use for the \"related\" command (defaults to 0)", "num");

		opts.add("invert-match", 'v', "invert-match", "invert the sense of matching, to select non-matching lines");
		opts.add("quiet", 'q', "quiet", "do not write anything to standard output, but exit with 0 if any match is found");
		
		opts.add("verbose", 0, "verbose", "enable verbose output");
		opts.add("debug", 0, "debug", "enable debugging output (including verbose output)");

		opts.addCommand("update", (int)UPDATE);
		opts.addCommand("check", (int)CHECK);
		opts.addCommand("tagshow", (int)TAGSHOW);
		opts.addCommand("tagsearch", (int)TAGSEARCH);
		opts.addCommand("show", (int)SHOW);
		opts.addCommand("related", (int)RELATED);
		opts.addCommand("cat", (int)CAT);
		opts.addCommand("grep", (int)GREP);
		opts.addCommand("install", (int)INSTALL);
		opts.addCommand("mkpatch", (int)MKPATCH);

		// Process the commandline
		valid_command cmd = (valid_command)opts.parse(argc, argv);

		CommandlineArgs args(argc, argv);

		if (opts.get("verbose").defined())
			Environment::get().verbose(true);

		if (opts.get("debug").defined())
			Environment::get().debug(true);
		
		// Perform the correct operation
		switch (cmd)
		{
			// Output the full package tag database
			case CAT:
			{
				wantTagDatabase();
				TagcollSerializer serializer(stdout);
				if (opts.get("groupitems").defined())
				{
					ItemGrouper<string> grouper;
					readFullCollection(grouper);
					grouper.output(serializer);
				} else
					readFullCollection(serializer);

				break;
			}

			// grep [-v] [-q] <tag expression>
			// Output the lines of the full package tag database that match the
			// given tag expression
			case GREP:
			{
				wantTagDatabase();
				string expression = args.next();

				// Build the grep filter chain
				FilterChain<string> filters;
				ExpressionFilter filter;
				if (!filter.setExpression(expression))
					fatal_error("Error in tag expression `%.*s'.\n", PFSTR(expression));
				filters.appendFilter(&filter);

				if (opts.get("invert-match").defined())
					filter.setMatchType(ExpressionFilter::INVERTED);

				if (opts.get("quiet").defined())
					filters.setConsumer(new TagcollSink<string>());
				else
					filters.setConsumer(new TagcollSerializer(stdout));

				readFullCollection(filters);

				return filter.countMatched() > 0 ? 0 : 1;
			}

			// install [-v] [-q] <tag expression>
			// apt-get install the packages that match the given tag expression
			case INSTALL:
			{
				wantTagDatabase();
				string expression = args.next();

				// Build the grep filter chain
				FilterChain<string> filters;
				ExpressionFilter filter;
				if (!filter.setExpression(expression))
					fatal_error("Error in tag expression `%.*s'.\n", PFSTR(expression));
				filters.appendFilter(&filter);

				if (opts.get("invert-match").defined())
					filter.setMatchType(ExpressionFilter::INVERTED);

				ItemCollector selection;
				filter.setConsumer(&selection);
				readFullCollection(filters);

				const char* parms[selection.size() + 1];
				int i = 0;
				parms[i++] = APTGET;
				parms[i++] = "install";
				for (ItemCollector::iterator tag = selection.begin();
						tag != selection.end(); tag++)
					parms[i++] = tag->c_str();
				parms[i] = 0;
				execv(APTGET, (char* const*)parms);
				throw SystemException(errno, "Running apt-get");
			}

			// tagshow <tag>
			// Show the vocabulary informations about a tag\n"
			case TAGSHOW:
			{
				wantTagDatabase();
				string tag = args.next();

				Vocabulary voc;
				voc.read();
				
				Vocabulary::const_iterator i = voc.find(tag);
				if (i == voc.end())
				{
					verbose("Tag `%.*s' was not found in tag vocabulary\n", PFSTR(tag));
					return 1;
				}
				else
				{
					printVocabularyItem(i->first, i->second);
					return 0;
				}
			}
			// tagsearch <pattern [pattern [pattern [...]]]>
			// Show a summary of all tags matching the given patterns
			case TAGSEARCH:
			{
				wantTagDatabase();
				// Get the patterns to be matched
				vector<string> patterns;
				while (args.hasNext())
					patterns.push_back(args.next());

				if (patterns.empty())
				{
					error("No patterns provided in commandline\n");
					return 1;
				}

				Vocabulary voc;
				voc.read();

				map<string, Vocabulary::Record> matches;
				for (Vocabulary::const_iterator i = voc.begin();
						i != voc.end(); i++)
				{
					// Try to match all patterns
					bool matched = true;
					for (vector<string>::const_iterator p = patterns.begin();
							matched && p != patterns.end(); p++)
						if (!matchVocabularyItem(i->first, i->second, *p))
							matched = false;

					if (matched)
						matches.insert(make_pair(i->first, i->second));
				}

				for (map<string, Vocabulary::Record>::const_iterator i = matches.begin();
						i != matches.end(); i++)
				{
					printVocabularyItem(i->first, i->second);
					fputc('\n', stdout);
				}

				return matches.size() > 0 ? 0 : 1;
			}
			// show <pkg>
			// Call apt-cache show <pkg>, but add tag informations to the output.\n"
			case SHOW:
			{
				wantTagDatabase();
				string pkg = args.next();
				
				TagcollCatcher catcher(pkg);
				readUnmergedFullCollection(catcher);
				OpSet<string> ts = catcher.tags();
				
				PackageDB db;
				string rec = db.getPackageRecord(pkg);
				
				if (rec.size())
				{
					printf("%.*sTags: ", PFSTR(rec));
					for (OpSet<string>::const_iterator i = ts.begin();
							i != ts.end(); i++)
						if (i == ts.begin())
							printf("%.*s", PFSTR(*i));
						else
							printf(", %.*s", PFSTR(*i));
					printf("\n");
					return 0;
				} else {
					verbose("Package %.*s not found", PFSTR(pkg));
					return 1;
				}
			}
			// related <pkg1[,pkg2[,pkg2,[...]]]>
			// Show packages related to the specified ones
			case RELATED:
			{
				wantTagDatabase();
				int maxdist = 0;
				if (opts.get("distance").defined())
					maxdist = opts.get("distance").intVal();
				string pkg = args.next();

				InputMerger<string> merger;
				readUnmergedCollection(merger);

				// Split the items on commas
				string splititem;
				set<string> splititems;
				for (string::const_iterator c = pkg.begin();
						c != pkg.end(); c++)
					if (*c == ',')
					{
						if (!merger.hasItem(splititem))
						{
							error("Item \"%.*s\" does not exist in the collection\n", PFSTR(splititem));
							return 1;
						}
						splititems.insert(splititem);
						splititem = string();
					} else
						splititem += *c;
				if (!merger.hasItem(splititem))
				{
					error("Item \"%.*s\" does not exist in the collection\n", PFSTR(splititem));
					return 1;
				}
				splititems.insert(splititem);

				// Get the tagset as the intersection of the tagsets of all input items
				set<string>::const_iterator i = splititems.begin();
				OpSet<string> ts = merger.getTagsetForItem(*i);
				for (++i; i != splititems.end(); i++)
					ts = ts ^ merger.getTagsetForItem(*i);

				if (ts.empty())
				{
					if (splititems.size() > 1)
						fprintf(stderr, "The packages %.*s are unrelated: cannot find a barycenter to start computing relationships from.\n", PFSTR(pkg));
					else
						fprintf(stderr, "The package %.*s has no tags attached.\n", PFSTR(pkg));
					return 1;
				}

				PackageDB db;
				
				// Build a full TagCollection
				TagcollBuilder builder(handleMaker);
				merger.output(builder);
				TagCollection<int> coll = builder.collection();

				set<int> items = coll.getItemsForTagset(ts);

				if (maxdist)
				{
					// Get the related tagsets
					list< OpSet<string> > rel = coll.getRelatedTagsets(ts, maxdist);

					// Print the output list
					for (list< OpSet<string> >::const_iterator i = rel.begin();
							i != rel.end(); i++)
					{
						set<int> idxs = coll.getItemsForTagset(*i);
						for (set<int>::const_iterator j = idxs.begin();
								j != idxs.end(); j++)
							items.insert(*j);
					}
				}
				printItems(db, items);

				if (items.size() > 50 && maxdist == 0 && isatty(1))
				{
					string tags;
					for (OpSet<string>::const_iterator i = ts.begin();
							i != ts.end(); i++)
						if (i == ts.begin())
							tags += *i;
						else
							tags += "%2C" + *i;
					feedback("\nIt seems that this set of packages lacks tag information that could help to better distinguish package similarities.\nYou can help providing better tagging: just point your browser to http://debian.vitavonni.de/packagebrowser/?tags=%.*s\n", PFSTR(tags));
				}
				return 0;
			}
			// check <file>
			// Check that all the tags in the given tagged collection are
			// present in the tag vocabulary.  Checks the main database if no
			// file is specified
			case CHECK:
			{
				wantTagDatabase();
				string file;
				if (args.hasNext())
					file = args.next();
				else
					file = fn_tagdb;

				Vocabulary voc;
				voc.read();

				TagcollChecker checker(voc);
				readCollection(file, checker);

				if (checker.missingCount() > 0)
				{
					checker.report();
					return 1;
				}
				else
					return 0;
			}
			// update
			// Updates the package tag database (requires root)
			case UPDATE:
			{
				/// Update the system vocabulary
				feedback("Merging system vocabularies...\n");
				Vocabulary voc;
				voc.readVocabularies();
				voc.write();
				voc.writeImplications();
				voc.writeDerivedTags();
				
				/// Update the system tag database
				feedback("Downloading tag database...\n");

				// Read the database sources
				vector<string> sources = readSources();

				// Acquire the databases from the sources
				InputMerger<string> merger;
				int found;
				for (vector<string>::const_iterator i = sources.begin();
						i != sources.end(); i++)
				{
					feedback("Acquiring source %.*s\n", PFSTR(*i));

					try {
						string local = acquireSource(*i);
						if (access(local.c_str(), R_OK) == -1)
							warning("Source %.*s has problems: ignoring it\n", PFSTR(*i));
						else
						{
							// FIXME: the input file is gzipped: must be read with a GzipParserInput
							feedback("Reading data from %.*s\n", PFSTR(local));
							zreadCollection(local, merger);
							found++;
						}
					} catch (Exception& e) {
						warning("%s: %.*s.  Ignoring this source\n",
								e.type(), PFSTR(e.desc()));
					}
				}
				
				if (!found)
					fatal_error("Unable to use any data source (not even previously cached ones)");

				
				// Prepare the input filter chain
				FilterChain<string> filters;

				// Build the patch list
				PatchList patches;
				int patchCount = 0;
				DIR* dir = opendir(path_tagpatch_d);
				if (!dir)
					throw SystemException(errno, string("reading directory ") + path_tagpatch_d);
				while (struct dirent* d = readdir(dir))
				{
					if (d->d_name[0] != '.' &&
							d->d_name[strlen(d->d_name) - 1] != '~')
					{
						string fn = string(path_tagpatch_d) + '/' + d->d_name;
						if (access(fn.c_str(), R_OK) == 0)
						{
							readCollection(fn, patches);
							patchCount++;
						}
					}
				}
				closedir(dir);


				// Prepare applying the patches
				if (patchCount)
					filters.appendFilter(new ApplyPatches(patches));

				ImplicationList implications;
				DerivedTagList derivedTags;

				voc.outputImplications(implications);
				voc.outputDerivedTags(derivedTags);

				implications.pack();

				// Prepare packing the database data

				// Expand implications
				filters.appendFilter(new ApplyImplications(implications));
				// Remove derived tags computing them using the expanded tag set
				filters.appendFilter(new CompressDerivedTags(derivedTags));
				// Compress implications
				filters.appendFilter(new CompressImplications(implications));

				// Group the items before output
				ItemGrouper<string> grouper;
				filters.setConsumer(&grouper);
				merger.output(filters);
				
				// Output the merged collection to the target file
				string tmpdb = string(fn_tagdb) + ".tmp";
				FILE* out = fopen(tmpdb.c_str(), "wt");
				if (!out) throw FileException(errno, "opening " + tmpdb);
				TagcollSerializer serializer(out);
				grouper.output(serializer);
				fclose(out);

				// Perform atomic update of the tag database
				if (rename(tmpdb.c_str(), fn_tagdb) == -1)
					throw FileException(errno, "renaming " + tmpdb + " to " + fn_tagdb);

				feedback("Done.\n");
				break;
			}
			// mkpatch [filename]
			// Create a tag patch between the current tag database and the tag
			// collection [filename]
			case MKPATCH:
			{
				wantTagDatabase();
				string file = args.next();

				InputMerger<string> merger1;
				InputMerger<string> merger2;

				readFullCollection(merger1);
				readFullCollection(file, merger2);

				PatchList newpatches;
				newpatches.generate(merger1, merger2);

				TagcollSerializer output(stdout);
				newpatches.output(output);
				break;
			}
		}

		return 0;
	}
	catch (Exception& e)
	{
		fatal_error("%s: %.*s\n", e.type(), PFSTR(e.desc()));
		return 1;
	}
}

// vim:set ts=4 sw=4:
